#'@rdname Analogs
#'@title Downscaling using Analogs based on large scale fields.
#' 
#'@author J. Ramon, \email{jaume.ramon@bsc.es}
#'@author Ll. Lledó, \email{llorenc.lledo@ecmwf.int}
#'
#'@description This function performs a downscaling using Analogs. To compute 
#'the analogs given a coarse-scale field, the function looks for days with 
#'similar conditions in the historical observations. The coarse scale and 
#'observation data can be either global or regional. In the latter case, the 
#'region is defined by the user. In principle, the coarse and observation data 
#'should be of the same variable, although different variables can also be admitted. 
#'The analogs function will find the N best analogs based in Minimum Euclidean 
#'distance.
#'
#'The search of analogs must be done in the longest dataset posible, but might 
#'require high-memory computational resources. This is important since it is 
#'necessary to have a good representation of the possible states of the field in 
#'the past, and therefore, to get better analogs. The function can also look for 
#'analogs within a window of D days, but is the user who has to define that window. 
#'Otherwise, the function will look for analogs in the whole dataset. This function 
#'is intended to downscale climate prediction data (i.e., sub-seasonal, seasonal 
#'and decadal predictions) but can admit climate projections or reanalyses. It does 
#'not have constrains of specific region or variables to downscale.
#'@param exp an 's2dv_cube' object containing the experimental field on the 
#'large scale for which the analog is aimed. The element 'data' in the 's2dv_cube' 
#'object must have, at least, the dimensions latitude, longitude, start date and time.
#'The object is expect to be already subset for the desired large scale region.
#'@param obs an 's2dv_cube' object containing the observational field with the
#'target high-resolution scale. The element 'data' in the 's2dv_cube' object must have, 
#'at least, the dimensions latitude, longitude, start date and time. The object is 
#'expect to be already subset for the desired large scale region.
#'@param nanalogs an integer indicating the number of analogs to be searched
#'@param fun_analog a function to be applied over the found analogs. Only these options
#'are valid: "mean", "wmean", "max", "min", "median" or NULL. If set to NULL (default), 
#'the function returns the found analogs.
#'@param lat_dim a character vector indicating the latitude dimension name in the element
#''data' in exp and obs. Default set to "lat".
#'@param lon_dim a character vector indicating the longitude dimension name in the element
#''data' in exp and obs. Default set to "lon".
#'@param sdate_dim a character vector indicating the start date dimension name in the
#'element 'data' in exp and obs. Default set to "sdate".
#'@param time_dim a character vector indicating the time dimension name in the element
#''data' in exp and obs. Default set to "time". This is expected to have daily frequency.
#'@param grid_exp a character vector indicating the coarse grid to be passed to CDO. It 
#'must be a grid recognised by CDO or a NetCDF file. This information can also be passed 
#'via the element 'source_files' in exp. Default to NULL.
#'@param ncores an integer indicating the number of cores to use in parallel computation.
#'@import multiApply
#'@import CSTools
#'@importFrom s2dv InsertDim CDORemap
#'@importFrom FNN get.knnx
#'
#'@seealso \code{\link[s2dverification]{CDORemap}}
#'
#'@return An 's2dv_cube' object containing the dowscaled values. If fun_analog is set to NULL,
#'the output array in 'data' also contains the dimension 'analog' with the best analog days.
#'@examples
#'exp <- rnorm(15000)
#'dim(exp) <- c(member = 5, lat = 4, lon = 5, sdate = 5, time  = 30)
#'lon_exp <- 1:5
#'lat_exp <- 1:4
#'obs <- rnorm(27000)
#'dim(obs) <- c(lat = 12, lon = 15, sdate = 5, time  = 30)
#'lon_obs <- seq(0,6, 6/14)
#'lat_obs <- seq(0,6, 6/11)
#'exp <- s2dv_cube(data = exp, lat = lat_exp, lon = lon_exp)
#'obs <- s2dv_cube(data = obs, lat = lat_obs, lon = lon_obs)
#'attr(exp$lon,"first_lon") <- 1
#'attr(exp$lon,"last_lon") <- 5
#'attr(exp$lat,"first_lat") <- 1
#'attr(exp$lat,"last_lat") <- 4
#'attr(obs$lon,"first_lon") <- 0
#'attr(obs$lon,"last_lon") <- 6
#'attr(obs$lat,"first_lat") <- 0
#'attr(obs$lat,"last_lat") <- 6
#'downscaled_field <- Analogs(exp = exp, obs = obs, grid_exp = 'r360x180')
#'@export
Analogs <- function(exp, obs, nanalogs = 3, fun_analog = "mean", lat_dim = "lat", lon_dim = "lon", 
                    sdate_dim = "sdate", time_dim = "time", grid_exp = NULL, ncores = 1) {
  #-----------------------------------
  # Checkings
  #-----------------------------------
  # input exp and obs must be s2dv_cube objects
  if (!inherits(exp,'s2dv_cube')) {
    stop("Parameter 'exp' must be of the class 's2dv_cube'")
  }

  # input exp and obs must be s2dv_cube objects
  if (!inherits(obs,'s2dv_cube')) {
    stop("Parameter 'obs' must be of the class 's2dv_cube'")
  }

  if (!inherits(nanalogs, 'numeric')) {
    stop("Parameter 'nanalogs' must be of the class 'character'")
  }

  if (!inherits(fun_analog, 'character')) {
    stop("Parameter 'fun_analog' must be of the class 'character'")
  }

  if (!inherits(lat_dim, 'character')) {
    stop("Parameter 'lat_dim' must be of the class 'character'")
  }

  if (!inherits(lon_dim, 'character')) {
    stop("Parameter 'lon_dim' must be of the class 'character'")
  }

  if (!inherits(sdate_dim, 'character')) {
    stop("Parameter 'sdate_dim' must be of the class 'character'")
  }

  if (!inherits(time_dim, 'character')) {
    stop("Parameter 'time_dim' must be of the class 'character'")
  }

  # Do not allow synonims for lat (latitude), lon (longitude) and time (sdate) dimension names 
  if (is.na(match(lon_dim, names(dim(exp$data)))) | is.na(match(lon_dim, names(dim(obs$data))))) {
    stop("Missing longitude dimension in 'exp' and/or 'obs', or does not match the parameter ",
         "'lon_dim'")
  } 

  if (is.na(match(lat_dim, names(dim(exp$data)))) | is.na(match(lat_dim, names(dim(obs$data))))) {
    stop("Missing latitude dimension in 'exp' and/or 'obs', or does not match the parameter ",
         "'lat_dim'")
  }

  if (is.na(match(sdate_dim, names(dim(exp$data)))) | is.na(match(sdate_dim, names(dim(obs$data))))) {
    stop("Missing start date dimension in 'exp' and/or 'obs', or does not match the parameter ",
         "'sdate_dim'")
  }

  if (is.na(match(time_dim, names(dim(exp$data)))) | is.na(match(time_dim, names(dim(obs$data))))) {
    stop("Missing time dimension in 'exp' and/or 'obs', or does not match the parameter ",
         "'time_dim'")
  }

  if (is.null(exp$source_files) & is.null(grid_exp)) {
      stop("I do not know which the coarse grid is. This information can be passed within ",
           "'exp' with the element '$source_files' or via the parameter 'grid_exp'. ",
           "Parameter 'grid_exp' must be a character vector indicating the coarse grid to ",
           "be passed to CDO, and it must be a grid recognised by CDO or a NetCDF file.")
  }

  # Ensure we have enough data to interpolate from high-res to coarse grid
  if ((attr(obs[[lat_dim]],"first_lat") > attr(exp[[lat_dim]],"first_lat")) | 
      (attr(obs[[lat_dim]],"last_lat") < attr(exp[[lat_dim]],"last_lat")) | 
      (attr(obs[[lon_dim]],"first_lon") > attr(exp[[lon_dim]],"first_lon")) |
      (attr(obs[[lon_dim]],"last_lon") < attr(exp[[lon_dim]],"last_lon"))) {
    stop("There are not enough data in 'obs'. Please to add more latitudes or ",
         "longitudes.")
  }

  # Select a function to apply to the analogs selected for a given observation
  stopifnot(fun_analog %in% c("mean", "wmean", "max", "min", "median"))

  # Create window if user does not have it in obs$data
  if ( !("window" %in% names(dim(obs$data))) ) {
    nsdates <- dim(obs$data)[names(dim(obs$data)) == sdate_dim]
    ntimes <- dim(obs$data)[names(dim(obs$data)) == time_dim]
    window <- Apply(list(obs$data), target_dims = list(c(time_dim, sdate_dim)), 
                    fun = as.vector, output_dims = 'window')$output1
    obs$data <- InsertDim(obs$data, posdim = 1, lendim = nsdates * ntimes, name = "window")
    obs$data <- Apply(list(obs$data, window), target_dims = 'window', 
                      fun = function(x,y) x <- y)$output1
  }

  #-----------------------------------
  # Interpolate high-res observations to the coarse grid
  #-----------------------------------
  if (is.null(grid_exp)) { 
    grid_exp <- exp$source_files[1]  
  } 

  lonmin <- attr(exp[[lon_dim]],"first_lon")
  lonmax <- attr(exp[[lon_dim]],"last_lon")
  latmin <- attr(exp[[lat_dim]],"first_lat")
  latmax <- attr(exp[[lat_dim]],"last_lat")
  obs_interpolated <- Interpolation(exp = obs, target_grid = grid_exp, method_remap = "conservative",
                                    remap_region = c(lonmin, lonmax, latmin, latmax))
  lats <- obs[[lat_dim]]
  lons <- obs[[lon_dim]]

  # If after interpolating 'obs' data the coordinates do not match, the exp data is interpolated to
  # the same grid to force the matching
  if (!.check_coords(lat1 = obs_interpolated$lat, lat2 = exp[[lat_dim]], lon1 = obs_interpolated$lon,
                     lon2 = exp[[lon_dim]])) {
    exp_interpolated <- Interpolation(exp = exp, target_grid = grid_exp, method_remap = "conservative",
                                      remap_region = c(lonmin, lonmax, latmin, latmax))
  } else {
    exp_interpolated <- exp
  }

  #-----------------------------------
  # Reshape train and test
  #-----------------------------------
  res <- Apply(list(obs_interpolated$data, exp_interpolated$data, obs$data), 
               target_dims = list(c("window", sdate_dim, time_dim, lat_dim, lon_dim), 
               c(sdate_dim, time_dim, lat_dim, lon_dim), c("window", sdate_dim, time_dim, lat_dim, lon_dim)), 
               fun = function(tr, te, ob) .analogs(train = tr, test = te, obs_hres = ob, k = nanalogs, 
               fun_analog = fun_analog), ncores = ncores)$output1

  #test <- exp$data[1,1,,1,,,]
  #train <- obs_interpolated$data[,1,1,,,,]
  #obs_hres <- obs$data[,1,1,,,,]
  #test <- aperm(test,c(2,1,3,4))
  #train <- aperm(train,c(1,4,2,3,5))
  #obs_hres <- aperm(obs_hres,c(1,4,2,3,5))  

  res_s2dv <- suppressWarnings(s2dv_cube(data = res, lon = lons, lat = lats))
  
  return(res_s2dv)
}


# For each element in test, find the indices of the k nearest neigbhors in train
.analogs <- function(train, test, obs_hres, k, fun_analog) {
  # train, test, and obs_hres  dim: 4 dimensions sdate, time, lat and lon (in this order)

  # Number of lats/lons of the high-resolution data
  space_dims_hres <- dim(obs_hres)[c(4,5)]
  nsdates <-  dim(train)[2]
  ntimes <- dim(train)[3]

  # Reformat train and test as an array with (time, points)
  train <- apply(train, c(1,2,3), as.vector); names(dim(train))[1] <- "space"  
  test <- apply(test, c(1,2), as.vector); names(dim(test))[1] <- "space"
  obs_hres <- apply(obs_hres, c(1,2,3), as.vector); names(dim(obs_hres))[1] <- "space"

  # Here we do cross-validation: the start date considered is removed from the training
  analogs_ls <- lapply(1:nsdates, function(s) {
    lapply(1:ntimes, function(t) {
      tr <- apply(train[ , , -s, t], 1, as.vector); names(dim(tr))[1] <- "time"
      te <- test[ , s, t]
      te <- InsertDim(data = te, posdim = 1, lendim = 1, name = 'time'); names(dim(te))[2] <- "space"
      ob <- apply(obs_hres[ , , -s, t], 1, as.vector); names(dim(ob))[1] <- "time"

      knn.ind <- get.knnx(tr, te, k)
   
      dist <- knn.ind$nn.dist
      idx <- knn.ind$nn.index
      names(dim(idx)) <- c("time", "analog"); names(dim(dist)) <- c("time", "analog")

      analogs <- Apply(idx, margins = "analog", fun = function(an) { ob[ an, ] })$output1
       
      dim(analogs) <- c(space_dims_hres, analogs = k)
 
      if (!is.null(fun_analog)) {

        if (fun_analog == "wmean") {
          weight <- 1 / dist
          analogs <- apply(analogs, c(1,2), function(x) weighted.mean(x, weight))
        } else {
          analogs <- apply(analogs, c(1,2), fun_analog)
        }
      }
    })
  })
  
  if (is.null(fun_analog)) {
    analogs_arr <- array(unlist(analogs_ls), dim = c(space_dims_hres, analogs = k, ntimes, nsdates))
  } else {
    analogs_arr <- array(unlist(analogs_ls), dim = c(space_dims_hres, ntimes, nsdates))
  }

  return(analogs_arr)
}


