Intbc <- function(exp, obs, target_grid, int_method, bc_method, ncores = 1) {

  require(CSTools)

  # Input data must be an s2dv_cube object
  if (!inherits(exp,'s2dv_cube')) {
    stop("Parameter 'exp' must be of the class 's2dv_cube'")
  }

  if (!inherits(obs,'s2dv_cube')) {
    stop("Parameter 'obs' must be of the class 's2dv_cube'")
  }

  if (!inherits(int_method, 'character')) {
    stop("Parameter 'int_method' must be of the class 'character'")
  }
 
  if (!inherits(bc_method, 'character')) {
    stop("Parameter 'bc_method' must be of the class 'character'")
  }
 
  stopifnot(bc_method %in% c('sbc', 'cal', 'qm', 'simple_bias', 'calibration', 'quantile_mapping'))
  stopifnot(c('member','sdate') %in% names(dim(exp$data)))

  #if ((attr(obs$lat,"first_lat") < attr(exp$lat,"first_lat")) |
  #    (attr(obs$lat,"last_lat") > attr(exp$lat,"last_lat")) |
  #    (attr(obs$lon,"first_lon") < attr(exp$lon,"first_lon")) |
  #    (attr(obs$lon,"last_lon") > attr(exp$lon,"last_lon"))) {
  #  stop("There are not enough data in 'exp'. Please to add more latitudes and ",
  #       "longitudes.")
  #}

  lonmin <- attr(obs$lon,"first_lon")
  lonmax <- attr(obs$lon,"last_lon")
  latmin <- attr(obs$lat,"first_lat")
  latmax <- attr(obs$lat,"last_lat")

  exp_interpolated <- Interpolation(exp = exp, target_grid = target_grid, method_remap = int_method,
                                    remap_region = c(lonmin, lonmax, latmin, latmax))

  # If after interpolating 'exp' data the coordinates do not match, the obs data is interpolated to
  # the same grid to force the matching
  if (!.check_coords(lat1 = exp_interpolated$lat, lat2 = obs$lat, lon1 = exp_interpolated$lon, 
                     lon2 = obs$lon)) {
    obs_interpolated <- Interpolation(exp = obs, target_grid = target_grid, method_remap = int_method,
                                      remap_region = c(lonmin, lonmax, latmin, latmax))
  } else {
    obs_interpolated <- obs
  }

  #.check_coords
  if (bc_method == 'sbc' | bc_method == 'simple_bias') {
    if (dim(obs_interpolated$data)['sdate'] == 1) {
      warning('Simple Bias Correction should not be used with 
              only one observation. Returning NA. ')
    }
    res <- BiasCorrection(exp = exp_interpolated$data, 
                          obs = obs_interpolated$data) 
  } 
  else if (bc_method == 'cal' | bc_method == 'calibration') {
    if (dim(exp_interpolated$data)['member'] == 1) {
      stop('Calibration must not be used with
           only one ensemble member.')
    }
    res <- Calibration(exp = exp_interpolated$data, obs = obs_interpolated$data, 
                       cal.method = 'mse_min', ncores = ncores)
  }
  else if (bc_method == 'qm' | bc_method == 'quantile_mapping') {
    if (any(is.na(exp_interpolated$data))) {
      warning('Found NAs in "exp" data, either introduced with the initial object or ', 
              'by the interpolation method. Quantile Mapping method is likely to fail.')
    }
    res <- QuantileMapping(exp = exp_interpolated$data, obs = obs_interpolated$data, 
                           sample_dims = 'sdate', method = 'QUANT', ncores = ncores)
  }

  # Create an s2dv_cube object
  res_s2dv <- suppressWarnings(s2dv_cube(data = res, lon = exp_interpolated$lon, lat = exp_interpolated$lat))

  return(res_s2dv)
}




