Intlr <- function(exp, obs, target_grid, int_method, lr_method, predictors = NULL, 
                  large_scale_predictor_dimname = 'vars', loocv = FALSE, ncores = 1) {

  require(multiApply) 

  #-----------------------------------
  # Checkings
  #-----------------------------------
  # input exp and obs must be s2dv_cube objects
  if (!inherits(exp,'s2dv_cube')) {
    stop("Parameter 'exp' must be of the class 's2dv_cube'")
  }
 
  # input exp and obs must be s2dv_cube objects
  if (!inherits(obs,'s2dv_cube')) {
    stop("Parameter 'obs' must be of the class 's2dv_cube'")
  }

  if (!inherits(int_method, 'character')) {
    stop("Parameter 'int_method' must be of the class 'character'")
  }
 
  if (!inherits(lr_method, 'character')) {
    stop("Parameter 'lr_method' must be of the class 'character'")
  }

  if (!inherits(large_scale_predictor_dimname, 'character')) {
    stop("Parameter 'large_scale_predictor_dimname' must be of the class 'character'")
  }

  if (!inherits(loocv, 'logical')) {
    stop("Parameter 'loocv' must be set to TRUE or FALSE")
  }

  # sdate must be the time dimension in the input data
  stopifnot('sdate' %in% names(dim(exp$data)))
  stopifnot('sdate' %in% names(dim(obs$data))) 
  
  # checkings for the parametre 'predictors' 
  if (is.array(predictors)) {
    # ensure the predictor variable name matches the parametre large_scale_predictor_dimname
    stopifnot(large_scale_predictor_dimname %in% names(dim(predictors)))
    stopifnot('sdate' %in% names(dim(predictors)))
    stopifnot(dim(predictors)['sdate'] == dim(exp$data)['sdate']) 
  }

  #-----------------------------------
  # Interpolation
  #-----------------------------------
  if (lr_method != '4nn') {
  
    lonmin <- attr(obs$lon,"first_lon")
    lonmax <- attr(obs$lon,"last_lon")
    latmin <- attr(obs$lat,"first_lat")
    latmax <- attr(obs$lat,"last_lat")

    exp_interpolated <- Interpolation(exp = exp, target_grid = target_grid, method_remap = int_method,
                                      remap_region = c(lonmin, lonmax, latmin, latmax))

    # If after interpolating 'exp' data the coordinates do not match, the obs data is interpolated to
    # the same grid to force the matching
    if (!.check_coords(lat1 = exp_interpolated$lat, lat2 = obs$lat, 
                       lon1 = exp_interpolated$lon, lon2 = obs$lon)) {
      obs_interpolated <- Interpolation(exp = obs, target_grid = target_grid, method_remap = int_method,
                                        remap_region = c(lonmin, lonmax, latmin, latmax))
    } else {
      obs_interpolated <- obs
    }
  }
 
  #-----------------------------------
  # Linear regressions
  #-----------------------------------
  # Pointwise linear regression
  # Predictor: model data
  # Predictand: observations
  if (lr_method == 'basic') {
    predictor <- exp_interpolated$data
    predictand <- obs_interpolated$data
    
    target_dims_predictor <- 'sdate'
    target_dims_predictand <- 'sdate'
   
    lats <- obs_interpolated$lat
    lons <- obs_interpolated$lon
  } 

  # (Multi) linear regression with large-scale predictors
  # Predictor: passed through the parameter 'predictors' by the user. Can be model or observations
  # Predictand: model data
  else if (lr_method == 'large-scale') {  
    if (is.null(predictors)) {
      stop("The large-scale predictors must be passed through the parametre 'predictors'")
    }  
    predictand <- obs_interpolated$data
    predictor <- predictors
    
    target_dims_predictor <- c('sdate', large_scale_predictor_dimname)
    target_dims_predictand <- 'sdate'
  
    lats <- obs_interpolated$lat
    lons <- obs_interpolated$lon
  } 

  # Multi-linear regression with the four nearest neighbours
  # Predictors: model data
  # Predictand: observations
  else if (lr_method == '4nn') {
    warning("Interpolating model and observation data but not needed.")
    predictor <- find_nn(hres = obs, coar = exp, nn = 4)
    predictand <- obs$data
    
    target_dims_predictor <- c('sdate','nn')
    target_dims_predictand <- 'sdate'
  
    lats <- obs$lat
    lons <- obs$lon
  }

  else {
    stop(paste0(lr_method, " method is not implemented yet"))
  }

  # Apply the linear regressions
  res <- Apply(list(predictor, predictand), target_dims = list(target_dims_predictor, target_dims_predictand), 
               fun = .intlr, loocv = loocv, ncores = ncores)$output1
 
  names(dim(res))[1] <- 'sdate'

  # Reorder dimensions to match those of the input model data
  res <- .reorder_dims(arr_ref = exp$data, arr_to_reorder = res)

  #-----------------------------------
  # Create an s2dv_cube object
  #-----------------------------------
  res_s2dv <- suppressWarnings(s2dv_cube(data = res, lon = lons, lat = lats))

  return(res_s2dv)
}

#-----------------------------------
# Atomic function to generate and apply the linear regressions
#-----------------------------------
.intlr <- function(x, y, loocv) {

  tmp_df <- data.frame(x = x, y = y)

  # training
  lm1 <- train_lm(df = tmp_df, loocv = loocv)  

  # prediction
  res <- pred_lm(lm1 = lm1, df = tmp_df, loocv = loocv)

  #return(lm1)
  return(res)
 
}

#-----------------------------------
# Function to generate the linear regressions.
# Returns a list
#-----------------------------------
train_lm <- function(df, loocv) {

  # Remove columns containing only NA's
  df <- df[ , apply(df, 2, function(x) !all(is.na(x)))]

  if (loocv) {
   
    lm1 <- lapply(1:nrow(df), function(j) lm(df[-j,], formula = y ~ .))
 
  } else {

    lm1 <- list(lm(data = df, formula = y ~ .))
  }

  return(lm1)
}

#-----------------------------------
# Function to apply the linear regressions.
#-----------------------------------
pred_lm <- function(df, lm1, loocv) {

  if (loocv) {

    pred_vals <- sapply(1:nrow(df), function(j) predict(lm1[[j]], df[j,]))
  
  } else {
  
    pred_vals_ls <- lapply(lm1, predict, data = df)    
    pred_vals <- unlist(pred_vals_ls)
  } 
 
  return(pred_vals)
}

#-----------------------------------
# Function to find N nearest neighbours.
# 'hres' and 'coar' are s2dv_objects
#-----------------------------------
find_nn <- function(hres, coar, nn = 4) {

  require(abind)

  lats_hres <- hres$lat
  lons_hres <- hres$lon
  lats_coar <- coar$lat
  lons_coar <- coar$lon
  
  # Sort the distances from closest to furthest
  idx_lat <- as.array(sapply(lats_hres, function(x) order(abs(lats_coar - x))[1:nn]))
  idx_lon <- as.array(sapply(lons_hres, function(x) order(abs(lons_coar - x))[1:nn]))

  names(dim(idx_lat)) <- c('nn', 'lat')
  names(dim(idx_lon)) <- c('nn', 'lon')

  # obtain the values of the nearest neighbours 
  nearest <- Apply(list(coar$data, idx_lat, idx_lon), 
                   target_dims = list(c('lat','lon'),'lat','lon'), 
                   fun = function(x, y, z) x[y, z])$output1

  return(nearest)
}


